/******************************************************************************/
/*                                                                            */
/*    Filename: AH4Control.c                                                  */
/*    Date: 190928 (dev start), 191231 (dev end) Rev 1.40                     */
/*    File Version: See below                                                 */
/*    Author: Hirofumi Momose / JJ2PNX                                        */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/*    Notes:                                                                  */
/*    For PCB rev 1.40 or higher                                              */
/*    Please use asm version for board before 1.40                            */
/*     due to different hardware design                                       */
/*    Copyright (c) 2019 JJ2PNX                                               */
/*     http://www.momose.com/hirofumi/jj2pnx/atucont/                         */
/*                                                                            */
/******************************************************************************/
/* Rev $Id: AH4Control.c 16 2019-12-30 15:47:45Z momose $ */

/******************************************************************************/
/* INCLUDES                                                                   */
/******************************************************************************/
#include <xc.h>

/******************************************************************************/
/* CONFIGURATION Bits                                                         */
/******************************************************************************/
#pragma config CP = OFF
#pragma config CPD = OFF
#pragma config WRT = OFF
#pragma config DEBUG = OFF
#pragma config LVP = OFF
#pragma config BOREN = ON
#pragma config MCLRE = OFF
#pragma config WDTE = OFF
#pragma config PWRTE = ON
#pragma config FOSC = INTOSCIO

// For version identification
#pragma config IDLOC0 = 1   // Soft rev 1.40
#pragma config IDLOC1 = 40

/******************************************************************************/
/* Common definition                                                          */
/******************************************************************************/
#define _XTAL_FREQ      (8000000)           // 8MHz
#define TICK_MS         (4.0)
#define MS2TICK(X)      ((unsigned char)((X) / TICK_MS +0.5))

/******************************************************************************/
/* I/O port definition                                                        */
/******************************************************************************/
// Port A
#define AH4_START       (PORTAbits.RA6)     // (out) AH4 Start
#define AH4_KEY         (PORTAbits.RA5)     // (in)  AH4 Key
#define ATT1_MASK       (0b10000000)        // (out) ATT1 (5dB, 100W)
#define ATT1            (PORTAbits.RA7)
#define ATT2_MASK       (0b00010000)        // (out) ATT2 (5dB, 30W)
#define ATT2            (PORTAbits.RA4)

// Port B
#define LED_TX_MASK     (0b00000001)        // (out) TX LED
#define LED_TX          (PORTBbits.RB0)
#define LED_SWR15_MASK  (0b00000010)        // (out) SWR1.5 LED
#define LED_SWR15       (PORTBbits.RB1)
#define LED_SWR20_MASK  (0b00000100)        // (out) SWR2.0 LED
#define LED_SWR20       (PORTBbits.RB2)
#define LED_AUTO_MASK   (0b00001000)        // (out) AUTO LED
#define LED_AUTO        (PORTBbits.RB3)
#define LED_TUNE_MASK   (0b10000000)        // (out) TUNE LED
#define LED_TUNE        (PORTBbits.RB7)
#define nTXDET          (PORTBbits.RB4)     // (in) TX Detect (neg)
#define nSW_AUTO        (PORTBbits.RB5)     // (in) AUTO sw (neg)
#define nSW_TUNE        (PORTBbits.RB6)     // (in) TUNE sw (neg)
// Note: Port B input (txdet, sw) change will wake cpu from sleep

/******************************************************************************/
/* EEPROM                                                                     */
/******************************************************************************/
// Address
#define EEADR_WAIT_RELAY    (0)
#define EEADR_VSWR_THRS     (1)
#define EEADR_VFWD_THRS     (2)
#define EEADR_AUTO          (3)

// Default data
#define EEDATA_WAIT_RELAY   (0)     // WAIT = 100 ms
#define EEDATA_VSWR_THRS    (4)     // VSWR = 2.0
#define EEDATA_VFWD_THRS    (5)     // POWER = 5 watts
#define EEDATA_AUTO         (1)     // AUTO ON
// See *_conf_table for options

// 'factory default'
__EEPROM_DATA(EEDATA_WAIT_RELAY, \
              EEDATA_VSWR_THRS, \
              EEDATA_VFWD_THRS, \
              EEDATA_AUTO, 0xff, 0xff, 0xff, 0xff);

unsigned char eep_read(unsigned char addr)
{
    while(EECON1bits.WR) continue; // Wait if previous write is in-progress
    EEADR = addr;
    EECON1bits.EEPGD = 0;
    EECON1bits.RD = 1;
    return EEDATA;
}

void eep_write(unsigned char addr, unsigned char data)
{
    while(EECON1bits.WR) continue; // Wait if previous write is in-progress
    EEADR = addr;
    EEDATA = data;
    EECON1bits.EEPGD = 0;
    EECON1bits.WREN = 1;
    INTCONbits.GIE = 0;
    EECON2 = 0x55;
    EECON2 = 0xaa;
    EECON1bits.WR = 1;
    INTCONbits.GIE =1;
    EECON1bits.WREN = 0;
}

/******************************************************************************/
/* Measure VFWD, VREF and calc VSWR                                           */
/******************************************************************************/
#define ADREF_P         (5.0)       // A/D +Reference VCC 5V
#define ADREF_N         (5.0/11.0)  // A/D -Reference AGND 5V * 1k/(10k+1k)
#define VFWD_LSB        ((ADREF_P - ADREF_N) /256.0)
#define VFWD_VAL(X)     (unsigned char)(((X)/VFWD_LSB) +0.5)
#define VSWR_LSB        (0.1)
#define VSWR_VAL(X)     (unsigned char)(((X)/VSWR_LSB) +0.5)

// Global
unsigned char vfwd;        // Forward voltage (lsb=4.5455/256 V)
unsigned char vswr;        // VSWR (lsb=0.1, Max 25.5)

// Fwd rf power to vfwd conversion
// vfwd(dc) = sqrt (Pfwd / 50) / 15 * 22 * 2 * 1.41 * 100 / 210
//  where  / 50 is line impedance Zo
//         / 15 is current trans ratio (1:15)
//         * 22 is load resistance
//         * 2  is from C + M (Vc and Vm must be adjusted to same)
//         * 1.41 is peak/rms
//         * 100 / 210 is R divider (100k/(10k+100k+100k))
// 
// Following voltage does not include AGND offset (approx 0.455V).
// Reading from TP1/2-GND will indicate +0.455V higher from listed below.
// 200W -> 3.95V
// 150W -> 3.42V
// 100W -> 2.79V
//  55W -> 2.07V
//  45W -> 1.87V
//  20W -> 1.25V
//  15W -> 1.08V
//  10W -> 0.88V
//   8W -> 0.79V
//   7W -> 0.74V
//   6W -> 0.68V
//   5W -> 0.62V
//   4W -> 0.56V
//   3W -> 0.48V
//   2W -> 0.40V
//   1W -> 0.28V
// 0.5W -> 0.20V
// 0.3W -> 0.15V

#define VFWD_55W    VFWD_VAL(2.07)
#define VFWD_45W    VFWD_VAL(1.87)
#define VFWD_20W    VFWD_VAL(1.25)
#define VFWD_15W    VFWD_VAL(1.08)
#define VFWD_7W     VFWD_VAL(0.74)
#define VFWD_6W     VFWD_VAL(0.68)
#define VFWD_5W     VFWD_VAL(0.62)
#define VFWD_4W     VFWD_VAL(0.56)
#define VFWD_3W     VFWD_VAL(0.48)
#define VFWD_2W     VFWD_VAL(0.40)
#define VFWD_1W     VFWD_VAL(0.28)
#define VFWD_0W5    VFWD_VAL(0.20)
#define VFWD_0W3    VFWD_VAL(0.15)

void measure(void)
{
    unsigned short us_vfwd = 0, us_vref = 0;
    unsigned short us_num, us_den, us_vswr;
    unsigned char i;

    // Measure 16 times to get averaged
    // Each A/D takes approx 40us (20us sampling + 20us conversion)
    // 16 conversion of 2ch will take 1280us approx
    // Actually measured 1600us for entire loop
    for(i = 0; i < 16; i++){
        // Measure forward voltage
        ADCON0 = 0b01001001;            // Fosc/16(2us), Channel 1, ADON
        __delay_us(20);                 // Sample time
        ADCON0bits.GO = 1;              // Start A/D
        while(ADCON0bits.GO) continue;
        us_vfwd += ((unsigned short)ADRESH << 8) | (unsigned short)ADRESL;

        // Measure reflected voltage
        ADCON0 = 0b01000001;            // Fosc/16(2us), Channel 0, ADON
        __delay_us(20);                 // Sample time
        ADCON0bits.GO = 1;              // Start A/D
        while(ADCON0bits.GO) continue;
        us_vref += ((unsigned short)ADRESH << 8) | (unsigned short)ADRESL;
    }
    // ADCON0bits.ADON = 0;            // AD off (moved to just before sleep)

    us_vfwd = us_vfwd / 16;
    us_vref = us_vref / 16;

    vfwd = (unsigned char)(us_vfwd >> 2);
    if(us_vfwd & 0x0002 && vfwd < 255){
        vfwd ++;   // '+0.5' for a bit improving
    }

    // Calc VSWR
    // vswr = |(vfwd + vref) / (vfwd - vref)|;
    us_num = (us_vfwd + us_vref) * 10;  // x10 (lsb=0.1)
    us_den = (us_vfwd > us_vref) ? (us_vfwd - us_vref) : (us_vref - us_vfwd);

    if(us_den > 0){
        us_vswr =  us_num / us_den;
    } else {
        us_vswr = 0xffff;
    }

    if(us_vswr > 255){
        // 8bit width
        us_vswr = 255;
    }

    if(us_vswr < VSWR_VAL(1.0)){
        // vswr >= 1.0 at any time on transmitting
        us_vswr = VSWR_VAL(1.0);
    }

    if(vfwd >= VFWD_0W3){
        vswr = (unsigned char)us_vswr;
    } else {
        // vswr inaccurate due to too lower power
        vswr = 0;
    }
}

/******************************************************************************/
/* Config                                                                     */
/******************************************************************************/
struct conf_st {
    unsigned char wait_relay;   // Relay stable wait (lsb=4ms)
    unsigned char vswr_thrs;    // VSWR threshold for auto mode (lsb=0.1)
    unsigned char vfwd_thrs;    // VFWD threshold for auto mode (lsb=4.5455/256 V)
    unsigned char isauto;       // Auto mode is selected
} conf;

// WAIT table
const unsigned char wait_conf_table[8] = {
    MS2TICK(100), MS2TICK(200), MS2TICK(300), MS2TICK(400),
    MS2TICK(500), MS2TICK(600), MS2TICK(700), MS2TICK(800)
};

// VSWR table
const unsigned char vswr_conf_table[8] = {
    VSWR_VAL(1.2), VSWR_VAL(1.4), VSWR_VAL(1.6), VSWR_VAL(1.8),
    VSWR_VAL(2.0), VSWR_VAL(2.2), VSWR_VAL(2.4), VSWR_VAL(2.6)
};
// Note:
// Change start of vswr table from 1.2 (older version start from 1.4)

// VFWD table
const unsigned char vfwd_conf_table[8] = {
    VFWD_0W5, VFWD_1W, VFWD_2W, VFWD_3W,
    VFWD_4W, VFWD_5W, VFWD_6W, VFWD_7W
};

/******************************************************************************/
/* ATT                                                                        */
/******************************************************************************/
enum att_e {
    ATT_OFF = 0,                    // ATT OFF
    ATT_5dB,                        // ATT  5dB, 100W (ATT1 only)
    ATT_10dB                        // ATT 10dB, 100W (ATT1, ATT2)
} att_current = ATT_OFF;

void att_select(enum att_e att)
{
    unsigned char tmp = PORTA;
    tmp &= ~(ATT1_MASK | ATT2_MASK);
    switch(att){
        case ATT_5dB:       // ATT1 (5dB, 100W)
            tmp |=  ATT1_MASK;
            break;

        case ATT_10dB:      // ATT1 (5dB, 100W) ->  ATT2(5dB, 30W)
            tmp |= (ATT1_MASK | ATT2_MASK);
            break;

        case ATT_OFF:       // ATT OFF
            break;

        default:
            // Unintended value, keep current state
            tmp = PORTA;
            break;
    }
    PORTA = tmp;
    att_current = att;
}
// Note:
// Special consideration is required for PIC device which does not have LATx
// register. Writing PORTx register, actually consists of read-modify-write.
// This 'read' reflects physical port state nonetheless of its direction.
// Writing port command sequentially and instanounsly could lead unintended
// results. Use temp register to avoid this issue.

/******************************************************************************/
/* ATU Control                                                                */
/******************************************************************************/
// ATU state
enum atu_state_e {
    ATU_UNTUNED = 0,                // Untuned (default state, manually untuned)
    ATU_TUNE_IN_PROGRESS,           // Tune in progress
    ATU_TUNED,                      // Tuned
    ATU_TUNE_FAILED                 // Tune failed (resulting in untuned)
} atu_state = ATU_UNTUNED;

// Control sequencer state
enum ctrl_state_e {
    CTRL_NONE = 0,                  // No operation (default state)
    CTRL_WAIT_RELAY_1,              // Wait for ALC stability prior to measure TX power
    CTRL_SELECT_ATT,                // Select corresponding ATT
    CTRL_WAIT_RELAY_2,              // Wait for ALC stability prior to start ATU tuning
    CTRL_ATU_START,                 // ATU start
    CTRL_ATU_WAIT_KEY,              // Wait for ATU ack
    CTRL_ATU_TUNE_IN_PROGRESS,      // ATU tune in progress
    CTRL_ATU_CHK_RESULT,            // Check ATU result
    CTRL_ATU_UNTUNE_IN_PROGRESS,    // ATU untune start(for RX)
    CTRL_INTERLUDE                  // Short delay prior to next operation
} ctrl_state = CTRL_NONE;

unsigned char ctrl_elapsed;         // Inter phase elapsed timer

void ctrl_start_tune(void)
{
    if(ctrl_state == CTRL_NONE){
        ctrl_elapsed = 0;
        att_select(ATT_10dB);
        ctrl_state = CTRL_WAIT_RELAY_1;
        atu_state = ATU_TUNE_IN_PROGRESS;
    }
}

void ctrl_start_untune(void)
{
    if(ctrl_state == CTRL_NONE){
        ctrl_elapsed = 0;
        AH4_START = 1;
        ctrl_state = CTRL_ATU_UNTUNE_IN_PROGRESS;
    }
}

void ctrl_abort_tune(void)
{
    switch(ctrl_state){
        case CTRL_NONE:
            break;
        case CTRL_WAIT_RELAY_1:
        case CTRL_SELECT_ATT:
        case CTRL_WAIT_RELAY_2:
        case CTRL_ATU_START:
        case CTRL_ATU_WAIT_KEY:
        case CTRL_ATU_TUNE_IN_PROGRESS:
        case CTRL_ATU_CHK_RESULT:
            AH4_START = 0;
            atu_state = ATU_UNTUNED;
            ctrl_state = CTRL_INTERLUDE;
            break;
        case CTRL_ATU_UNTUNE_IN_PROGRESS:
        case CTRL_INTERLUDE:
        default:
            break;
    }
}

void ctrl_task(void)
{
    switch(ctrl_state){

       case CTRL_WAIT_RELAY_1:
           if(ctrl_elapsed++ < conf.wait_relay) break;
           ctrl_elapsed = 0;
           ctrl_state = CTRL_SELECT_ATT;
           // NO break required (thru)

       case CTRL_SELECT_ATT:
           ctrl_elapsed = 0;
           ctrl_state = CTRL_WAIT_RELAY_2;
           if(vfwd <= VFWD_15W) att_select(ATT_OFF);        //   ~15W ATT OFF
           else if(vfwd <= VFWD_45W) att_select(ATT_5dB);   // 15~45W ATT 5dB
           else ctrl_state = CTRL_ATU_START;                // 45~    ATT 10dB
           // ATT not change and additional wait not required when power > 45W
           break;

       case CTRL_WAIT_RELAY_2:
           if(ctrl_elapsed++ < conf.wait_relay) break;
           ctrl_elapsed = 0;
           ctrl_state = CTRL_ATU_START;
           // NO break required (thru)

       case CTRL_ATU_START:
           ctrl_elapsed = 0;
           AH4_START = 1;
           ctrl_state = CTRL_ATU_WAIT_KEY;
           break;

       case CTRL_ATU_WAIT_KEY:
           ctrl_elapsed = 0;
           if(AH4_KEY == 1) break;
           AH4_START = 0;
           ctrl_state = CTRL_ATU_TUNE_IN_PROGRESS;
           // No break required (thru)

       case CTRL_ATU_TUNE_IN_PROGRESS:
           // Over power protection
           // Insert att immmediately if power goes out of expected range
           if( att_current == ATT_OFF && vfwd >= VFWD_20W
            || att_current == ATT_5dB && vfwd >= VFWD_55W ){
               att_select(ATT_10dB);
           }
           if(AH4_KEY == 0) break;
           ctrl_elapsed = 0;
           ctrl_state = CTRL_ATU_CHK_RESULT;
           break;

       case CTRL_ATU_CHK_RESULT:
           if(ctrl_elapsed++ < MS2TICK(50)) break;
           ctrl_elapsed = 0;
           if(AH4_KEY == 1) atu_state = ATU_TUNED;
           else atu_state = ATU_TUNE_FAILED;
           ctrl_state = CTRL_INTERLUDE;
           break;

       case CTRL_ATU_UNTUNE_IN_PROGRESS:
           if(ctrl_elapsed++ < MS2TICK(50)) break;
           ctrl_elapsed = 0;
           AH4_START = 0;
           atu_state = ATU_UNTUNED;
           ctrl_state = CTRL_INTERLUDE;
           break;

       case CTRL_INTERLUDE:
           if(ctrl_elapsed++ < MS2TICK(100)) break;
           ctrl_elapsed = 0;
           ctrl_state = CTRL_NONE;
           break;

       case CTRL_NONE:
       default:
           AH4_START = 0;
           att_select(ATT_OFF);
           break;
    }
}

/******************************************************************************/
/* Push SW                                                                    */
/******************************************************************************/
#define BOUNCE_DELAY       MS2TICK(20)    // Bounce ignore time (20ms)

union {
    unsigned char uc;
    struct {
        unsigned auto_edge_en :1;
        unsigned tune_edge_en :1;
        unsigned dummy :6;
    };
} sw_flags = {0xff};

unsigned char sw_auto_count = 255;
unsigned char sw_tune_count = 255;

#define SW_INIT() \
do { \
    sw_auto_count = 255; \
    sw_tune_count = 255; \
    sw_flags.uc = 0xff; \
} while(0)

void sw_poll(void)
{
    if(nSW_AUTO){
        // SW released (pull-up)
        if(sw_auto_count < 255) sw_auto_count ++;
        if(sw_auto_count >= BOUNCE_DELAY){
            sw_flags.auto_edge_en = 1;
        }
    } else {
        // SW pressed
        sw_auto_count = 0;
    }

    if(nSW_TUNE){
        // SW released (pull-up)
        if(sw_tune_count < 255) sw_tune_count ++;
        if(sw_tune_count >= BOUNCE_DELAY){
            sw_flags.tune_edge_en = 1;
        }
    } else {
        // SW pressed
        sw_tune_count = 0;
    }
}

unsigned char sw_auto_edge(void)
{
    if(sw_auto_count < BOUNCE_DELAY){
        if(sw_flags.auto_edge_en ){
            sw_flags.auto_edge_en = 0;
            return 1;
        }
    }
    return 0;
}

unsigned char sw_tune_edge(void)
{
    if(sw_tune_count < BOUNCE_DELAY){
        if(sw_flags.tune_edge_en){
            sw_flags.tune_edge_en= 0;
            return 1;
        }
    }
    return 0;
}

unsigned char sw_tune_level(void)
{
    if(sw_tune_count < BOUNCE_DELAY) return 1;
    return 0;
}

/******************************************************************************/
/* LED                                                                        */
/******************************************************************************/
void led_disp(void)
{
    static unsigned char blinker;
    unsigned char tmp = PORTB;
    
    // Turn off all led
    tmp &= ~(LED_TX_MASK | LED_SWR15_MASK | LED_SWR20_MASK | LED_AUTO_MASK | LED_TUNE_MASK);

#if 1
    if(vswr >= VSWR_VAL(1.0)){
        // vswr >= 1.0 when transmitting
        if(vfwd >= conf.vfwd_thrs){
            // RF power exceeds threshold
            tmp |= LED_TX_MASK;
        } else {
            // Blinking when rf power does not exceed
            tmp |= (blinker & 0x40) ? LED_TX_MASK : 0;
        }
        if(vswr >= VSWR_VAL(1.5)) tmp |= LED_SWR15_MASK;
        if(vswr >= VSWR_VAL(2.0)) tmp |= LED_SWR20_MASK;
    }
#endif

    if(conf.isauto) tmp |= LED_AUTO_MASK;

    switch(atu_state){
        case ATU_TUNED:
            tmp |= LED_TUNE_MASK;
            break;
        case ATU_TUNE_IN_PROGRESS:
            tmp |= (blinker & 0x40) ? LED_TUNE_MASK : 0;
            break;
        default:
            break;
    }

    blinker ++;
    PORTB = tmp;
}

void led_num2bin(unsigned char num)
{
    unsigned char tmp = PORTB;
    tmp &= ~(LED_TX_MASK | LED_SWR15_MASK | LED_SWR20_MASK);
    if(num & 0x01) tmp |= LED_SWR20_MASK;
    if(num & 0x02) tmp |= LED_SWR15_MASK;
    if(num & 0x04) tmp |= LED_TX_MASK;
    PORTB = tmp;
}

/******************************************************************************/
/* Interrupt handler                                                          */
/******************************************************************************/
#define TMR0CLK         (_XTAL_FREQ /4 /64)
#define MS2TMR0(X)      ((unsigned char)((X) *TMR0CLK /1000.0 +0.5))
union {
    unsigned char uc;
    struct {
        unsigned xtsk_4ms :1;
        unsigned dummy :7;
    };
} isr_flags = { 0 };
#define xtsk_4ms    (isr_flags.xtsk_4ms)

void interrupt isr(void)
{
    // Periodic timer interrupt (4ms)
    if(INTCONbits.TMR0IF){
        INTCONbits.TMR0IF = 0;
        TMR0 = -MS2TMR0(TICK_MS);
        xtsk_4ms = 1;
    }

    // Port B Change interrupt
    // Invoked when
    //  TX detect (RB4)
    //  Push/release auto sw (RB5)
    //  Push/release tune sw (RB6)
    if(INTCONbits.RBIF){
        // PORTB read or write is required to end mismatch condition
        // that continue to set RBIF
        volatile unsigned char tmp = PORTB; // Dummy read prior to clear the flag
        INTCONbits.RBIF = 0;
        // Do nothing (wakeup from sleep only)
        // All input handling is processed in the main loop
    }
}

/******************************************************************************/
/* Main                                                                       */
/******************************************************************************/
void main(void)
{
    unsigned char met_elapsed = 0, rx_elapsed = 255;

    // OSC setup
    OSCCON = 0b01110000; // Select 8MHz

    // Setup port A
    TRISA = 0b00101111; // b7: ATT1 (out)
                        // b6: AH4 Start (out)
                        // b5: AH4 Key (in)
                        // b4: ATT2 (out)
                        // b3: A/D VRef + (analog)
                        // b2: A/D VRef - (analog)
                        // b1: A/D VFwd (analog)
                        // b0: A/D VRef (analog)
    PORTA = 0;

    // Setup port B
    TRISB = 0b01110000; // b7: LED Tune (out)
                        // b6: SW  Tune (in)
                        // b5: SW  Auto (in)
                        // b4: TX det (in)
                        // b3: LED Auto (out)
                        // b2: LED SWR 2.0 (out)
                        // b1: LED SWR 1.5 (out)
                        // b0: LED TX det (out)
    PORTB = 0;
    OPTION_REGbits.nRBPU = 0;	// Enable Port B pull-up

    // Setup A/D
    ADCON1 = 0b11001101;// Right justified, A/D Clock div by 2
                        // AN3: VREF+  (VCC)
                        // AN2: VREF-  (AGND)
                        // AN1: Analog (Fwd pwr)
                        // AN0: Analog (Ref pwr)

    // Setup timer0
    OPTION_REGbits.T0CS = 0;    // Use Fosc/4
    OPTION_REGbits.PSA = 0;     // Use prescaler
    OPTION_REGbits.PS = 0b101;  // Set prescaler 1/64

    // Setup interrupt
    INTCONbits.TMR0IE = 1;      // Enable timer0 interrupt
    INTCONbits.RBIE = 1;        // Enable Port B Change interrupt
    INTCONbits.GIE = 1;         // Enable interrupt

    // Load config from EEPROM
    conf.wait_relay= wait_conf_table[ eep_read(EEADR_WAIT_RELAY) ];
    conf.vswr_thrs = vswr_conf_table[ eep_read(EEADR_VSWR_THRS) ];
    conf.vfwd_thrs = vfwd_conf_table[ eep_read(EEADR_VFWD_THRS) ];
    conf.isauto = eep_read(EEADR_AUTO);

    __delay_ms(100);

    if(nSW_AUTO == 0 && nSW_TUNE == 0){
        // Pressing both AUTO and TUNE at POR
        // Restore eeprom with 'factory default'
        LED_AUTO = 1;
        eep_write(EEADR_WAIT_RELAY, EEDATA_WAIT_RELAY);
        eep_write(EEADR_VSWR_THRS,  EEDATA_VSWR_THRS);
        eep_write(EEADR_VFWD_THRS,  EEDATA_VFWD_THRS);
        eep_write(EEADR_AUTO,       EEDATA_AUTO);
        __delay_ms(100);
        LED_TUNE = 1;
        while(1) continue;
        // Power reset to exit
    }

    if(nSW_AUTO == 0){
        // Pressing only AUTO at POR
        // Config mode
        unsigned char isvalue = 0, target = 0 , value;
        LED_AUTO = 1;
        while(nSW_AUTO == 0) continue;
        __delay_ms(100);
        LED_AUTO = 0;

        while(1){
            if(xtsk_4ms){
                // 4ms periodic
                xtsk_4ms = 0;
                sw_poll();

                if(sw_auto_edge()){
                    // Toggle target/value
                    isvalue ^= 1;
                    value = eep_read(target);
                }

                if(sw_tune_edge()){
                    if(isvalue){
                        // Change value and save to eeprom
                        if(value < 7) value++;
                        else value = 0;
                        eep_write(target, value);
                    } else {
                        // Change target
                        if(target < 2) target ++;
                        else target = 0;
                    }
                }

                LED_TUNE = sw_tune_level();
                if(isvalue){
                    LED_AUTO = 1;
                    led_num2bin(value);
                } else {
                    LED_AUTO = 0;
                    led_num2bin(target);
                }
            }
        }
        // Power reset to exit
    }

    while(1){
        // Normal control loop
        if(xtsk_4ms){
            // 4ms periodic
            xtsk_4ms = 0;

            // Poll switch state
            sw_poll();

            // Auto-sw
            if(sw_auto_edge()){
                conf.isauto ^= 0x01;                // Toggle auto-mode
                conf.isauto &= 0x01;
                eep_write(EEADR_AUTO, conf.isauto); // Save to eeprom
            }

            if(nTXDET == 0){
                // Transmitting
                rx_elapsed = 0;
            } else {
                // Recieving
                if(rx_elapsed < 255) rx_elapsed ++;
            }

            if(rx_elapsed < MS2TICK(50)){
                // Transmitting w/rx delay 50ms
                measure();  // Measure and calc vswr
                
                // Start tune manually
                if(sw_tune_edge()) ctrl_start_tune();

                if(ctrl_state == CTRL_INTERLUDE && vfwd >= conf.vfwd_thrs
                 && (atu_state == ATU_TUNE_FAILED
                  || (atu_state == ATU_TUNED && vswr >= conf.vswr_thrs))){
                    // Turn off auto-mode if tune failed or 
                    //  tuned with vswr exceeds threshold
                    // This will prevent such loop that
                    //  high vswr -> start tuning ->
                    //  failed (or tuned with high vswr) -> high vswr ...
                    conf.isauto = 0;
                    // This is temporaliry off, no eeprom save
                }

                if(conf.isauto && ctrl_state == CTRL_NONE
                  && vfwd >= conf.vfwd_thrs && vswr >= conf.vswr_thrs){
                    // Full auto and both rf power and vswr exceeds threshold
                    if(met_elapsed++ > MS2TICK(50)){
                        // Start tune automatically when condition met for 50ms
                        ctrl_start_tune();
                        met_elapsed = 0;
                    }
                } else {
                    met_elapsed = 0;
                }

            } else {
                // Receiving
                vswr = 0;               // Set vswr = 0 to turn off vswr led
                ctrl_abort_tune();      // Cancel ongoing tuning process

                // Start untune
                if(sw_tune_edge()) ctrl_start_untune();
            }

            ctrl_task();
            led_disp();
            //led_num2bin(vfwd);
#if 1
            if(ctrl_state == CTRL_NONE && rx_elapsed > MS2TICK(100)
             && sw_auto_count > MS2TICK(100) && sw_tune_count > MS2TICK(100)){
                // Prepare for sleep
                att_select(ATT_OFF);
                AH4_START = 0;
                ADCON0bits.ADON = 0;    // AD off
                // CPU sleep to eliminate noise emission during recieving
                SLEEP();
                // CPU will wake when Port B input (Tx det, SW1, SW2) changes
                rx_elapsed = 255;
                SW_INIT();
            }
#endif
        } // if(xtsk_4ms)
    } // while(1)
    // Not reach
} // main
// EOF